package reksten.base;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import neustore.base.DBRuntimeException;

public abstract class DataObject {

	public static final short FLOAT_VECTOR = 1;
	
	/**
     * Creates a new copy of the object.
     * @return  the new copy
     */
    public abstract Object value();
    
    /** 
     * Returns the number of bytes the object occupy.
     * @return  number of bytes
     */
    public abstract int size();
    
    /**
     * Returns the maximum number of bytes this type of object may occupy.
     * It is used to support variable-length data.
     * @return  maximum number of bytes
     */
    public abstract int maxSize();
    
    /**
     * Reads the object from an input stream.
     * @param in   input stream
     * @throws IOException
     */
    public abstract void read(DataInputStream in) throws IOException;
    
    /**
     * Writes the object to an output stream.
     * @param out   output stream
     * @throws IOException
     */
    public abstract void write(DataOutputStream out) throws IOException;

    public static DataObject parse(DataInputStream in) throws IOException {
    	
    	short value = in.readShort();
    	if(value==FLOAT_VECTOR) {
    		FloatVector object = new FloatVector();
    		object.read(in);
    		return object;
    	}
    	
    	throw new DBRuntimeException("Object not instantiable!");
    }
	
}
