package reksten.base;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

public class FloatVector extends DataObject {

	static int MAX_SIZE = 100*4;
	private float[] vector;
	
	public FloatVector() {
		this.vector = new float[] {};
	}
	
	public FloatVector(float[] list) {
		this.vector = new float[list.length];
		System.arraycopy(list,0,this.vector,0,list.length);
	}
	
	public int length() {
		return this.vector.length;
	}
	
	public float get(int i) {
		return this.vector[i];
	}
	
	@Override
	public int maxSize() {
		return FloatVector.MAX_SIZE;
	}
	
	public int dimensionality() {
		return this.vector.length;
	}

	@Override
	public int size() {
		return 2+4+4*vector.length;
	}

	@Override
	public Object value() {
		float[] arr = new float[this.vector.length];
		System.arraycopy(this.vector,0,arr,0,arr.length);
		return arr;
	}
	
	@Override
	public void read(DataInputStream in) throws IOException {
		
		int len = in.readInt();
		this.vector = new float[len];
		for(int i = 0;i<len;i++)this.vector[i] = in.readFloat();

	}

	@Override
	public void write(DataOutputStream out) throws IOException {
		out.writeShort(DataObject.FLOAT_VECTOR);
		out.writeInt(this.vector.length);
		for(float val : this.vector)out.writeFloat(val);

	}
	
	public String toString() {
		return Arrays.toString(vector);
	}
	
	public boolean equals(FloatVector flv) {
		if(this.vector.length!=flv.vector.length)return false;
		for(int i = 0;i<this.vector.length;i++) {
			if(!Helper.equals(this.vector[i],flv.vector[i]))return false;
		}
		return true;
	}
	
	public FloatVector copy() {
		return new FloatVector(this.vector);
	}

}
