package reksten.linearscan;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

import reksten.base.DataObject;

import neustore.base.DBPage;
import neustore.base.DBRuntimeException;

public class LScanPage<D extends DataObject> extends DBPage {

	// Page ID of no page
	public static final int NO_PAGE = -1;
	
	// Page Types
	public static final int DATA_PAGE = 1;
	
	// Number of bytes used by the page header:
	private static final int HEADER_SIZE = 40;
	
	// Private variables
	private int previous;
	private int freeBytes;
	private int numRecs;
	private LinkedList<D> datalist;
	
	public LScanPage(int nodeType, int pageSize, int previousPage) {
		super(nodeType, pageSize);
		this.previous = previousPage;
		freeBytes = pageSize - HEADER_SIZE;
		numRecs = 0;
		datalist = new LinkedList<D>();
	}

	public int previous() {
		return this.previous;
	}
	
	public boolean enoughSpace(D d) {
		return d.size()<=freeBytes;
	}
	
	public Iterator<D> iterator() {
		return this.datalist.iterator();
	}
	
	public void insert(D d) {
		if(!enoughSpace(d))throw new DBRuntimeException("Page full!");
		datalist.add(d);
		freeBytes -= d.size();
		numRecs++;
	}
	
	/**
	 * Reads the object from a byte array.
	 * 
	 * @param page
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected void read(byte[] page) throws IOException {
		ByteArrayInputStream byte_in = new ByteArrayInputStream(page);
        DataInputStream in = new DataInputStream(byte_in);
		this.nodeType = in.readInt();
		this.pageSize = in.readInt();
		this.previous = in.readInt();
		this.freeBytes = in.readInt();
		this.numRecs = in.readInt();
		this.datalist = new LinkedList<D>();
		for(int i = 0;i<numRecs;i++) {
			D data = (D)DataObject.parse(in);
			datalist.add(data);
		}
		in.close();
		byte_in.close();
	}

	/**
	 * Writes the object to a byte array.
	 * 
	 * @param page
	 * @throws IOException
	 */
	@Override
	protected void write(byte[] page) throws IOException {
		ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
		DataOutputStream out = new DataOutputStream(byte_out);
		
		out.writeInt(this.nodeType);
		out.writeInt(this.pageSize);
		out.writeInt(this.previous);
		out.writeInt(this.freeBytes);
		out.writeInt(this.numRecs);
		
		for(D data : datalist)data.write(out);
		
		byte[] result = byte_out.toByteArray();
		System.arraycopy(result, 0, page, 0, result.length );
		
		out.close();
		byte_out.close();
	}
	
	
	
}
