package reksten.pmtree;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import reksten.base.DataObject;

import neustore.base.DBRuntimeException;

public abstract class PMTreeEntry<Data extends DataObject> {

	static final short NODE_LEAF = 0;
	static final short NODE_INTERNAL = 1;
	
	protected Data object;
	protected short nodeType = 0;
	protected float parentdistance;
	
	@SuppressWarnings("unchecked")
	public static PMTreeEntry parse(DataInputStream in) throws IOException {
		
		short nodeType = in.readShort();
		
		if(nodeType==NODE_LEAF) {
			PMTreeLeafEntry leaf = new PMTreeLeafEntry();
			leaf.read(in);
			return leaf;
		} else if(nodeType==NODE_INTERNAL) {
			PMTreeInternalEntry internal = new PMTreeInternalEntry();
			internal.read(in);
			return internal;
		}
		
		throw new DBRuntimeException("Node type does not exist!");
	}
	/**
	 * 
	 * @return The amount of bytes this element occupies.
	 */
	public abstract int size();
	
	public float getParentDistance() {
		return this.parentdistance;
	}
	
	public abstract float getRadius();
	
	public Data getObject() {
		return this.object;
	}
	
	public abstract void updateHyperRings(PMTreeInternalEntry<Data> parent);
	
	protected abstract void read(DataInputStream in) throws IOException;

	protected void write(DataOutputStream out) throws IOException {
		out.writeShort(this.nodeType);
	}
	
}
