package reksten.pmtree;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

import reksten.base.DataObject;


public class PMTreeInternalEntry<Data extends DataObject> extends PMTreeEntry<Data> {

	private int subtreeid;				// Page ID for subtree	
	private float radius;				// Radius of covering ball
	protected float[] hyperrings_inner;	// Lower bound on distance from static (LAESA) pivots
	protected float[] hyperrings_outer;	// Upper bound on distance from static (LAESA) pivots
	
	public PMTreeInternalEntry() {
		this.nodeType = PMTreeEntry.NODE_INTERNAL;
	}
	
	public PMTreeInternalEntry(PMTree<Data> tree, PMTreePage<Data> subpage, Data pivot, int pageid) {
		this.nodeType = PMTreeEntry.NODE_INTERNAL;
		this.subtreeid = pageid;
		this.object = pivot;
		this.radius = subpage.calculateRadius(tree, object);
		subpage.updateHyperrings(this);
	}
	
	public void update(PMTree<Data> tree, PMTreePage<Data> subpage) {
		this.radius = subpage.calculateRadius(tree, object);
		subpage.updateHyperrings(this);
	}
	
	public float[] getHyperring(int dimension) {
		return new float[] {hyperrings_inner[dimension],hyperrings_outer[dimension]};
	}
	
	public int size() {
		int sz = object.size();
		sz += 2; // this.nodeType;
		sz += 4; // subtreeid
		sz += 4; // radius
		sz += 4; // parentdistance
		sz += 4; // length of hyperrings tables
		sz += 4*hyperrings_inner.length; // inner hyperrings
		sz += 4*hyperrings_outer.length; // outer hyperrings
		return sz;
	}
	
	public float getRadius() {
		return this.radius;
	}
	
	public int getSubtreeID() {
		return this.subtreeid;
	}
	
	@Override
	public void updateHyperRings(PMTreeInternalEntry<Data> parent) {
		if(parent.hyperrings_inner==null) {
			parent.hyperrings_inner = new float[this.hyperrings_inner.length];
			Arrays.fill(parent.hyperrings_inner, Float.MAX_VALUE);
			parent.hyperrings_outer = new float[this.hyperrings_outer.length];
		}
		for(int i = 0;i<hyperrings_inner.length;i++) {
			parent.hyperrings_inner[i] = Math.min(parent.hyperrings_inner[i], this.hyperrings_inner[i]);
			parent.hyperrings_outer[i] = Math.max(parent.hyperrings_outer[i], this.hyperrings_outer[i]);
		}
		
	}
	
	@SuppressWarnings("unchecked")
	protected void read(DataInputStream in) throws IOException {
		object = (Data)DataObject.parse(in);
		subtreeid = in.readInt();
		radius = in.readFloat();
		parentdistance = in.readFloat();
		int len = in.readInt();
		hyperrings_inner = new float[len];
		for(int i = 0;i<len;i++)hyperrings_inner[i] = in.readFloat();
		hyperrings_outer = new float[len];
		for(int i = 0;i<len;i++)hyperrings_outer[i] = in.readFloat();
	}

	protected void write(DataOutputStream out) throws IOException {
		super.write(out);
		object.write(out);
		out.writeInt(subtreeid);
		out.writeFloat(radius);
		out.writeFloat(parentdistance);
		out.writeInt(hyperrings_inner.length);
		for(float d : hyperrings_inner)out.writeFloat(d);
		for(float d : hyperrings_outer)out.writeFloat(d);
	}

	
	
}
