package reksten.pmtree;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

import reksten.base.DataObject;


public class PMTreeLeafEntry<Data extends DataObject> extends PMTreeEntry<Data> {

//	private Data data;			// Data object stored in this node
//	private float parentdistance;		// Distance from data object to parent pivot
	private float[] pivotdistances;	// Distances from data object to static (LAESA) pivots
	
	public PMTreeLeafEntry() {
		this.nodeType = PMTreeEntry.NODE_LEAF;
	}
	
	public PMTreeLeafEntry(Data object, float parentdistance, float[] pivotdistances) {
		this.nodeType = PMTreeEntry.NODE_LEAF;
		this.object = object;
		this.parentdistance = parentdistance;
		this.pivotdistances = new float[pivotdistances.length];
		System.arraycopy(pivotdistances,0,this.pivotdistances,0,this.pivotdistances.length);
	}
	
	public int size() {
		int sz = object.size();
		sz += 2; // this.nodeType;
		sz += 4; // parentdistance
		sz += 4; // length of pivotdistance table
		sz += 4 * pivotdistances.length; // inner hyperrings
		return sz;
	}
	
	public float getRadius() {
		return 0.0f;
	}
	
	public float getPivotDistance(int index) {
		return this.pivotdistances[index];
	}

	@SuppressWarnings("unchecked")
	protected void read(DataInputStream in) throws IOException {
		this.object = (Data)DataObject.parse(in);
		this.parentdistance = in.readFloat();
		this.pivotdistances = new float[in.readInt()];
		for(int i = 0;i<this.pivotdistances.length;i++)
			pivotdistances[i] = in.readFloat();
	}

	protected void write(DataOutputStream out) throws IOException {
		super.write(out);
		object.write(out);
		out.writeFloat(parentdistance);
		out.writeInt(pivotdistances.length);
		for(float d : pivotdistances)out.writeFloat(d);
	}

	@Override
	public void updateHyperRings(PMTreeInternalEntry<Data> parent) {
		if(parent.hyperrings_inner==null) {
			parent.hyperrings_inner = new float[this.pivotdistances.length];
			Arrays.fill(parent.hyperrings_inner, Float.MAX_VALUE);
			parent.hyperrings_outer = new float[this.pivotdistances.length];
		}
		for(int i = 0;i<parent.hyperrings_inner.length;i++) {
			parent.hyperrings_inner[i] = Math.min(parent.hyperrings_inner[i], this.pivotdistances[i]);
			parent.hyperrings_outer[i] = Math.max(parent.hyperrings_outer[i], this.pivotdistances[i]);
		}
	}
	
}
