package reksten.tests;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.StringTokenizer;

import reksten.base.Distance;
import reksten.base.FloatVector;
import reksten.base.QFDistance;
import reksten.base.SimSearchIndex;
import reksten.linearscan.LScan;
import reksten.pmtree.PMTree;
import reksten.xtree.XTree;

import neustore.base.DBBuffer;
import neustore.base.LRUBuffer;

public class TestRunner {

	static final float MIN = -100.0f;
	static final float MAX = 100.0f;
	
	static final float EPSILON = 0.1f;
	static final int PIVOTS = 10;
	static final float MIN_FANOUT = 0.35f;
	static final float MAX_OVERLAP = 0.2f;
	
	static final int BUFFER_SIZE = 2048;
	static final int PAGE_SIZE = 8192;
	
	static final float RANGE = 10.0f;
	static final int K = 40;
	
	static final int TESTRUNS = 50;
	
	static BufferedReader stdin;
	static StringTokenizer st;

	static String LINE() throws Exception { return stdin.readLine(); }
	static String TOKEN() throws Exception {
		while (st == null || !st.hasMoreTokens())st = new StringTokenizer(LINE());
		return st.nextToken();
	}
	static int INT() throws Exception {return Integer.parseInt(TOKEN());}
	static long LONG() throws Exception {return Long.parseLong(TOKEN());}
	static float FLOAT() throws Exception {return Float.parseFloat(TOKEN());}
	
	static DecimalFormat DF = new DecimalFormat("0.00000",new DecimalFormatSymbols(Locale.ENGLISH));
	
	public static float[] parseLine(int N) throws Exception {
		float[] ret = new float[N];
		for(int i = 0;i<N;i++) {
			ret[i] = FLOAT();
		}
		return ret;
	}
	
	public static void main(String[] args) throws Exception {
		
		
		
		dimensionToQFIndex = new HashMap<Integer, Integer>();
		dimensionToQFIndex.put(2,0);
		dimensionToQFIndex.put(4,0);
		dimensionToQFIndex.put(6,0);
		dimensionToQFIndex.put(8,0);
		dimensionToQFIndex.put(12,0);
		dimensionToQFIndex.put(16,0);
		dimensionToQFIndex.put(24,0);
		dimensionToQFIndex.put(32,0);

		
		String[] TEST_DATA = new String[] {
			"YOGA2D.test",
			"YOGA4D.test",
			"YOGA8D.test",
			"YOGA12D.test",
			"YOGA16D.test",
			"YOGA24D.test",
			"YOGA32D.test",
		};
		
		int[] xaxis = new int[] {2,4,8,12,16,24,32};
		
		for(int i = 1;i<3;i++) {
			
			BufferedWriter distancesearch = new BufferedWriter(new FileWriter(new File("plotYOGA"+i+"distancesearch.plot")));
			BufferedWriter disksearch = new BufferedWriter(new FileWriter(new File("plotYOGA"+i+"disksearch.plot")));
			BufferedWriter cpusearch = new BufferedWriter(new FileWriter(new File("plotYOGA"+i+"cpusearch.plot")));
			BufferedWriter distanceinsert = new BufferedWriter(new FileWriter(new File("plotYOGA"+i+"distanceinsert.plot")));
			BufferedWriter diskinsert = new BufferedWriter(new FileWriter(new File("plotYOGA"+i+"diskinsert.plot")));
			BufferedWriter cpuinsert = new BufferedWriter(new FileWriter(new File("plotYOGA"+i+"cpuinsert.plot")));
			
			for(int test = 0;test<TEST_DATA.length;test++) {
				
				System.err.println("****************************************");
				System.err.println("************** GO ["+i+", "+test+"]****************");
				System.err.println("****************************************");
				
				stdin = new BufferedReader(new FileReader(new File(TEST_DATA[test])));

				int N = INT(), D = INT();
				
				long[][] rr = null;
				if(i==0)rr = doLinearTest(N,D);
				if(i==1)rr = doPMTreeTest(N,D);
				if(i==2)rr = doPtolemaicPMTreeTest(N, D);
				if(i==3)rr = doXTreeTest(N, D);
				
				distancesearch	.write(xaxis[test]+" "+DF.format(rr[1][0]/(float)TESTRUNS)+"\n");
				disksearch		.write(xaxis[test]+" "+DF.format(rr[1][1]/(float)TESTRUNS)+"\n");
				cpusearch		.write(xaxis[test]+" "+DF.format(rr[1][2]/(float)TESTRUNS/1000.0)+"\n");
				distanceinsert	.write(xaxis[test]+" "+DF.format(rr[2][0]/(float)TESTRUNS)+"\n");
				diskinsert		.write(xaxis[test]+" "+DF.format(rr[2][1]/(float)TESTRUNS)+"\n");
				cpuinsert		.write(xaxis[test]+" "+DF.format(rr[2][2]/(float)TESTRUNS/1000.0)+"\n");
				
				stdin.close();
			}
			distancesearch.flush(); distancesearch.close();
			disksearch.flush(); disksearch.close();
			cpusearch.flush(); cpusearch.close();
			distanceinsert.flush(); distanceinsert.close();
			diskinsert.flush(); diskinsert.close();
			cpuinsert.flush(); cpuinsert.close();
		}
		
		
		// *****************************************************************************
		// ************************** TESTING CALL *************************************
		// *****************************************************************************
		
		
		
		
	}
	
	public static long[][] doLinearTest(int N, int D) throws Exception {
		
		if(!dimensionToQFIndex.containsKey(D))throw new IllegalArgumentException("Missing Quadratic Form!");
		int qfindex = dimensionToQFIndex.get(D);
		
		// 1. Create the index!
		DBBuffer buffer = new LRUBuffer(BUFFER_SIZE, PAGE_SIZE);
		Distance<FloatVector> dist = new QFDistance(quadtraticforms[qfindex], EPSILON);
		
		LScan<FloatVector> index = new LScan<FloatVector>(buffer, "linear.test", true, dist);
		for(int i = 0;i<N;i++) {
			index.insert(new FloatVector(parseLine(D)));
		}
		
		long[][] res = performTests(index,buffer,dist, D);
		printResults(res);
		System.err.println();
		pgfPlotResults(D,TESTRUNS,res);
		return res;
	}
	
	public static long[][] doPMTreeTest(int N, int D) throws Exception {
		
		if(!dimensionToQFIndex.containsKey(D))throw new IllegalArgumentException("Missing Quadratic Form!");
		int qfindex = dimensionToQFIndex.get(D);
		
		// 1. Create the index!
		DBBuffer buffer = new LRUBuffer(BUFFER_SIZE, PAGE_SIZE);
		Distance<FloatVector> dist = new QFDistance(quadtraticforms[qfindex], EPSILON);
		
		List<FloatVector> pivots = new ArrayList<FloatVector>();
		for(int i = 0; i<PIVOTS;i++) {
			pivots.add(new FloatVector(randomList(MIN, MAX, D)));
		}
		
		PMTree<FloatVector> index = new PMTree<FloatVector>(buffer, "pmtree.test", PAGE_SIZE, pivots, dist, false);
		for(int i = 0;i<N;i++) {
//			System.err.println(i);
			index.insert(new FloatVector(parseLine(D)));
		}
		long[][] res = performTests(index,buffer,dist, D);
		printResults(res);
		System.err.println();
		pgfPlotResults(D,TESTRUNS,res);
		return res;
	}
	
	public static long[][] doPtolemaicPMTreeTest(int N, int D) throws Exception {
		
		if(!dimensionToQFIndex.containsKey(D))throw new IllegalArgumentException("Missing Quadratic Form!");
		int qfindex = dimensionToQFIndex.get(D);
		
		// 1. Create the index!
		DBBuffer buffer = new LRUBuffer(BUFFER_SIZE, PAGE_SIZE);
		Distance<FloatVector> dist = new QFDistance(quadtraticforms[qfindex], EPSILON);
		
		List<FloatVector> pivots = new ArrayList<FloatVector>();
		for(int i = 0; i<PIVOTS;i++) {
			pivots.add(new FloatVector(randomList(MIN, MAX, D)));
		}
		
		PMTree<FloatVector> index = new PMTree<FloatVector>(buffer, "ppmtree.test", PAGE_SIZE, pivots, dist, true);
		for(int i = 0;i<N;i++) {
//			if(i%100==99)System.err.println((i+1));
			index.insert(new FloatVector(parseLine(D)));
		}
		
		long[][] res = performTests(index,buffer,dist, D);
		printResults(res);
		System.err.println();
		pgfPlotResults(D,TESTRUNS,res);
		return res;
	}
	
	public static long[][] doXTreeTest(int N, int D) throws Exception {
		
		if(!dimensionToQFIndex.containsKey(D))throw new IllegalArgumentException("Missing Quadratic Form!");
		int qfindex = dimensionToQFIndex.get(D);
		
		// 1. Create the index!
		DBBuffer buffer = new LRUBuffer(BUFFER_SIZE, PAGE_SIZE);
		Distance<FloatVector> dist = new QFDistance(quadtraticforms[qfindex], EPSILON);
		
		List<FloatVector> pivots = new ArrayList<FloatVector>();
		for(int i = 0; i<PIVOTS;i++) {
			pivots.add(new FloatVector(randomList(MIN, MAX, D)));
		}
		
		XTree<FloatVector> index = new XTree<FloatVector>(buffer, "xtree.test", true, PAGE_SIZE, MIN_FANOUT, MAX_OVERLAP, dist);
		for(int i = 0;i<N;i++) {
			index.insert(new FloatVector(parseLine(D)));
		}
		
		long[][] res = performTests(index,buffer,dist, D);
		printResults(res);
		System.err.println();
		pgfPlotResults(D,TESTRUNS,res);
		return res;
	}
	
	public static void pgfPlotResults(int D, int N, long[][] res) {
		System.err.println("Range");
		System.err.println(D+" "+(res[0][0]/(double)N));
		System.err.println(D+" "+(res[0][1]/(double)N));
		System.err.println(D+" "+(res[0][2]/(double)N));
		System.err.println("Nearest Neighbor");
		System.err.println(D+" "+(res[1][0]/(double)N));
		System.err.println(D+" "+(res[1][1]/(double)N));
		System.err.println(D+" "+(res[1][2]/(double)N));
		System.err.println("Insert");
		System.err.println(D+" "+(res[2][0]/(double)N));
		System.err.println(D+" "+(res[2][1]/(double)N));
		System.err.println(D+" "+(res[2][2]/(double)N));
	}
	
	public static void printResults(long[][] res) {
		System.err.println("Range Searches:");
		System.err.println("  Distance Calculations => "+res[0][0]);
		System.err.println("  Disk Accesses         => "+res[0][1]);
		System.err.println("  CPU Time              => "+res[0][2]);
		
		System.err.println("Nearest Neighbor Searches");
		System.err.println("  Distance Calculations => "+res[1][0]);
		System.err.println("  Disk Accesses         => "+res[1][1]);
		System.err.println("  CPU Time              => "+res[1][2]);
		
		System.err.println("Insertions");
		System.err.println("  Distance Calculations => "+res[2][0]);
		System.err.println("  Disk Accesses         => "+res[2][1]);
		System.err.println("  CPU Time              => "+res[2][2]);
	}
	
	public static long[][] performTests(SimSearchIndex<FloatVector> index, DBBuffer buffer, Distance<FloatVector> dist, int dimensions) throws Exception {
		
		System.err.println("Start tests!");
		
		long startdistancecalcs = dist.calculations();
		long startdiskaccesses = buffer.getIOs()[0]+buffer.getIOs()[1];
		long[][] results = new long[3][3];
		long time_start = 0;
		// 1. Do Range Search tests
		time_start = System.currentTimeMillis();
		for(int i = 0;i<TESTRUNS;i++)
			index.rangeSearch(new FloatVector(randomList(MIN, MAX, dimensions)), RANGE);
		results[0][0] = dist.calculations();
		results[0][1] = buffer.getIOs()[1]+buffer.getIOs()[0];
		results[0][2] = System.currentTimeMillis()-time_start;
		// 2. Do K-nearest neighbor search tests
		time_start = System.currentTimeMillis();
		for(int i = 0;i<TESTRUNS;i++)
			index.kNearestNeighborSearch(new FloatVector(randomList(MIN, MAX, dimensions)), K);
		results[1][0] = dist.calculations() - results[0][0];
		results[1][1] = buffer.getIOs()[1]+buffer.getIOs()[0] - results[0][1];
		results[1][2] = System.currentTimeMillis()-time_start;
		// 3. Do insertion tests
		time_start = System.currentTimeMillis();
		for(int i = 0;i<TESTRUNS;i++) {
			index.insert(new FloatVector(randomList(MIN, MAX, dimensions)));
		}
		results[2][0] = dist.calculations() - results[1][0]-results[0][0];
		results[2][1] = buffer.getIOs()[1]+buffer.getIOs()[0] - results[1][1] - results[0][1];
		results[2][2] = System.currentTimeMillis()-time_start;
		
		results[0][0] -= startdistancecalcs;
		results[0][1] -= startdiskaccesses;
		
		return results;
		
	}
	
	public static float[] randomList(float min, float max, int D) {
		float[] ret = new float[D];
		for(int i = 0;i<D;i++)
			ret[i] = (float)(min+(max-min)*Math.random());
		return ret;
	}
	
	static HashMap<Integer,Integer> dimensionToQFIndex;
	
	static final float[][][] quadtraticforms = new float[][][] {
		new float[][] {	{4.8689904f, 0.44246244f},
						{0.44246244f, 0.48409283f}},
		new float[][] {	{0.055662613f, -0.21012987f, -0.1968944f, -0.19253755f},
						{-0.21012987f, 0.95435035f, -1.2920741f, -1.7184857f},
						{-0.1968944f, -1.2920741f, 3.1140285f, 1.9566162f},
						{-0.19253755f, -1.7184857f, 1.9566162f, 4.168807f}},
		new float[][] {	{4.955774f, -0.95523f, 0.46867847f, -0.38101077f, 1.0295291f, -0.44221592f},
						{-0.95523f, 1.8782163f, -1.3558099f, 0.5139291f, -1.2575455f, 0.031951904f},
						{0.46867847f, -1.3558099f, 3.7118871f, -1.134039f, -0.68344426f, 1.3035312f},
						{-0.38101077f, 0.5139291f, -1.134039f, 2.0630543f, 0.7794056f, -0.4769206f},
						{1.0295291f, -1.2575455f, -0.68344426f, 0.7794056f, 4.6052675f, -3.1937046f},
						{-0.44221592f, 0.031951904f, 1.3035312f, -0.4769206f, -3.1937046f, 4.887488f}},
		new float[][] {	{1.1542817f, -1.4308162f, 0.2017138f, 0.0831002f, 0.6753055f, 0.1331523f, 0.06567642f, 0.38361785f},
						{-1.4308162f, 3.641369f, -0.36523914f, 0.27115226f, -1.2501334f, -0.2633474f, -0.34526238f, 0.084193945f},
						{0.2017138f, -0.36523914f, 2.8896646f, -0.29787087f, 0.09587884f, 0.21999097f, -0.04226619f, -0.22346961f},
						{0.0831002f, 0.27115226f, -0.29787087f, 2.2484787f, -0.36943603f, -0.079559326f, -0.16542974f, -0.39828378f},
						{0.6753055f, -1.2501334f, 0.09587884f, -0.36943603f, 1.0983069f, 0.08133471f, 0.24457842f, 0.21002889f},
						{0.1331523f, -0.2633474f, 0.21999097f, -0.079559326f, 0.08133471f, 0.63339454f, -0.071986616f, 0.20031583f},
						{0.06567642f, -0.34526238f, -0.04226619f, -0.16542974f, 0.24457842f, -0.071986616f, 0.101471454f, -0.018843718f},
						{0.38361785f, 0.084193945f, -0.22346961f, -0.39828378f, 0.21002889f, 0.20031583f, -0.018843718f, 0.13633287f}},
		new float[][] { {0.8666292f, -0.48007584f, -0.3204918f, -1.8554529f, 0.5235452f, 0.34005678f, -0.6893389f, -0.05849567f, -0.061348796f, -1.6609496f, 0.19818324f, 0.09789443f},
						{-0.48007584f, 0.74923974f, -0.6843028f, 0.021279216f, 0.34394956f, 0.1185703f, 0.66365564f, -0.30551252f, -0.38674527f, 0.35113502f, 0.37083215f, 1.076788f},
						{-0.3204918f, -0.6843028f, 3.560242f, 0.89841413f, -0.9435665f, -0.3879126f, 0.024041176f, 0.48691857f, 1.6087372f, 1.047257f, -1.1069069f, 2.2835047f},
						{-1.8554529f, 0.021279216f, 0.89841413f, 4.3887157f, -0.16118693f, 0.26355016f, 0.8235717f, 0.27723402f, 1.0747802f, 3.27779f, -0.86797035f, 3.2847445f},
						{0.5235452f, 0.34394956f, -0.9435665f, -0.16118693f, 0.6878929f, 0.06234345f, -0.8562533f, 0.078469336f, -0.11181486f, -1.590846f, -0.26044434f, 1.3637567f},
						{0.34005678f, 0.1185703f, -0.3879126f, 0.26355016f, 0.06234345f, 0.18117678f, 0.48872852f, -0.17023884f, -0.20766607f, -0.8249403f, 0.20302376f, -0.1675244f},
						{-0.6893389f, 0.66365564f, 0.024041176f, 0.8235717f, -0.8562533f, 0.48872852f, 1.7709428f, -0.42030254f, 1.2918224f, 1.2431028f, 0.05558324f, -1.1892297f},
						{-0.05849567f, -0.30551252f, 0.48691857f, 0.27723402f, 0.078469336f, -0.17023884f, -0.42030254f, 0.16783449f, 0.4300347f, -0.045674562f, -0.35665622f, -0.44519544f},
						{-0.061348796f, -0.38674527f, 1.6087372f, 1.0747802f, -0.11181486f, -0.20766607f, 1.2918224f, 0.4300347f, 2.20206f, 2.850022f, -0.35324335f, 0.41138196f},
						{-1.6609496f, 0.35113502f, 1.047257f, 3.27779f, -1.590846f, -0.8249403f, 1.2431028f, -0.045674562f, 2.850022f, 4.565111f, -0.31902838f, 4.031784f},
						{0.19818324f, 0.37083215f, -1.1069069f, -0.86797035f, -0.26044434f, 0.20302376f, 0.05558324f, -0.35665622f, -0.35324335f, -0.31902838f, 0.8817611f, 0.020834327f},
						{0.09789443f, 1.076788f, 2.2835047f, 3.2847445f, 1.3637567f, -0.1675244f, -1.1892297f, -0.44519544f, 0.41138196f, 4.031784f, 0.020834327f, 3.6447942f}},
		new float[][] {	{9.6263485f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.09492779f, 0.0f, 0.0f, 1.7032061f, 0.0f, 0.0f, -0.5963745f},
						{0.0f, 6.25636f, 0.94123745f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.5317364f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.94123745f, 4.866194f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 9.673691f, 0.0f, 0.0f, 1.4910498f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.7406759f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 5.6605144f, 0.0f, 0.0f, 1.1661043f, 0.0f, 0.43091345f, 0.33585978f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 6.4898896f, 0.0f, 0.0f, 0.0f, 1.0794001f, 0.10074186f, -0.3677292f, 0.0f, 0.0f, 0.351274f, 0.0f},
						{0.0f, 0.0f, 0.0f, 1.4910498f, 0.0f, 0.0f, 6.800181f, 0.0f, -0.76953554f, 0.0f, 0.0f, -1.713469f, 0.0f, 0.50065804f, -0.49965715f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 1.1661043f, 0.0f, 0.0f, 8.547688f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.76953554f, 0.0f, 7.894742f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{-0.09492779f, 0.0f, 0.0f, 0.0f, 0.43091345f, 1.0794001f, 0.0f, 0.0f, 0.0f, 9.77766f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.33585978f, 0.10074186f, 0.0f, 0.0f, 0.0f, 0.0f, 8.438107f, 0.0f, 0.0f, 0.0f, 0.0f, -1.0804725f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.3677292f, -1.713469f, 0.0f, 0.0f, 0.0f, 0.0f, 8.722527f, -3.4427624f, 0.0f, 0.83958244f, 0.0f},
						{1.7032061f, 0.5317364f, 0.0f, -0.7406759f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -3.4427624f, 9.429132f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.50065804f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 5.0980353f, 0.18909836f, -0.86768675f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.351274f, -0.49965715f, 0.0f, 0.0f, 0.0f, 0.0f, 0.83958244f, 0.0f, 0.18909836f, 7.59958f, 0.0f},
						{-0.5963745f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.0804725f, 0.0f, 0.0f, -0.86768675f, 0.0f, 5.95447f}},
		new float[][] {	{9.166118f, 0.0f, 1.2070403f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 4.4064407f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.07493591f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.81685305f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{1.2070403f, 0.0f, 6.621512f, 1.2507591f, -1.6419082f, -1.0719433f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 1.2507591f, 5.369982f, 0.0f, -1.350975f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.1788206f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, -1.6419082f, 0.0f, 9.888418f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, -1.0719433f, -1.350975f, 0.0f, 8.558149f, 0.0f, -1.0045118f, 0.0f, 0.0f, 1.5404787f, 0.0f, 0.0f, 0.0f, 0.0f, -0.4509964f, 0.0f, 0.0f, 0.0f, 0.0f, 4.1531725f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 7.7647853f, 0.0f, -0.8298116f, 0.0f, 0.0f, 0.37965488f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, -0.07493591f, 0.0f, 0.0f, 0.0f, -1.0045118f, 0.0f, 5.1305285f, -0.3291707f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.3863144f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.8298116f, -0.3291707f, 7.109067f, -1.8566499f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.8566499f, 9.495794f, 0.0f, -0.93093395f, 0.0f, 0.0f, 0.0f, 1.4010391f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.5404787f, 0.0f, 0.0f, 0.0f, 0.0f, 8.498385f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.36235714f, 0.0f, -0.9228201f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.37965488f, 0.0f, 0.0f, -0.93093395f, 0.0f, 9.1575775f, 0.0f, 0.0f, -1.1793609f, 0.0f, -0.011253357f, 0.0f, 0.0f, -0.13836384f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 7.4475107f, 0.0f, 0.0031571388f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.16641426f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 5.6730156f, -0.62374973f, 0.0f, 0.0f, 0.0f, -0.43583918f, 0.0f, 0.8117957f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.1793609f, 0.0031571388f, -0.62374973f, 4.472069f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.9370928f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.4509964f, 0.0f, 0.3863144f, 0.0f, 1.4010391f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 7.5179334f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.011253357f, 0.0f, 0.0f, 0.0f, 0.0f, 7.192165f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.36235714f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 4.1567764f, 2.3338509f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.81685305f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.43583918f, 0.0f, 0.0f, 0.0f, 2.3338509f, 7.288493f, 0.0f, 0.0f, 0.0f, 0.0f, 0.690845f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.9228201f, -0.13836384f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 7.205518f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 1.1788206f, 0.0f, 4.1531725f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.8117957f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 9.821277f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.16641426f, 0.0f, -0.9370928f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 9.617345f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 5.734305f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.690845f, 0.0f, 0.0f, 0.0f, 0.0f, 8.105312f}},
		new float[][] {	{7.518986f, 1.0287762f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0724363f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{1.0287762f, 9.313089f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.43196774f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 9.075523f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 2.1248732f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 7.762816f, 0.0f, 0.0f, -1.1795783f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 9.656056f, 0.0f, 0.0f, 0.0f, 0.0f, 1.3080463f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 7.154087f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 2.0970755f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, -1.1795783f, 0.0f, 0.0f, 4.256258f, 0.0f, -1.7833452f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.7221575f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 6.8824606f, 0.6261902f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.4776812f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.7833452f, 0.6261902f, 8.918693f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.0495162f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.734354f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 1.3080463f, 0.0f, 0.0f, 0.0f, 0.0f, 8.473105f, 0.0f, 0.0f, -0.63170576f, 0.9356489f, -0.7418661f, 0.03406048f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 7.867296f, 0.535408f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.60257244f, 0.0f, -1.4427409f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.535408f, 6.8608685f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.1056509f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.63170576f, 0.0f, 0.0f, 5.717162f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{1.0724363f, 0.0f, 2.1248732f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.9356489f, 0.0f, 0.0f, 0.0f, 8.823101f, 0.0f, 0.0f, 1.2769055f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.7418661f, 0.0f, 0.0f, 0.0f, 0.0f, 6.736463f, -0.5874295f, 0.0f, 0.0f, -0.6423888f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.06886673f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.03406048f, 0.0f, 0.0f, 0.0f, 0.0f, -0.5874295f, 4.738638f, 0.0f, 0.0f, 0.0f, -0.800508f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.0495162f, 0.0f, 0.0f, 0.0f, 0.0f, 1.2769055f, 0.0f, 0.0f, 6.2363243f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.4617321f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.60257244f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 9.419215f, -0.534915f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.1056509f, 0.0f, 0.0f, -0.6423888f, 0.0f, 0.0f, -0.534915f, 6.2588377f, 0.0f, 0.0f, 0.0f, 0.0f, -1.0260525f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.4427409f, 0.0f, 0.0f, 0.0f, 0.0f, -0.800508f, 0.0f, 0.0f, 0.0f, 9.439047f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 5.9777393f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.55663586f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.4776812f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 4.6298857f, 0.8207507f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -0.7221575f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.8207507f, 5.163952f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.0260525f, 0.0f, 0.0f, 0.0f, 0.0f, 7.9059644f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 2.0970755f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.06886673f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 6.7371955f, 0.0f, 0.0f, 0.0f, 0.0f, 0.91752195f, 0.0f, 0.0f},
						{0.0f, -0.43196774f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.4617321f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 4.586616f, 1.5100207f, 0.0f, 0.0f, 0.0f, -1.3735518f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.734354f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.5100207f, 4.8460956f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 9.822611f, 0.0f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 4.2456665f, 0.0f, 0.0f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.55663586f, 0.0f, 0.0f, 0.0f, 0.91752195f, 0.0f, 0.0f, 0.0f, 0.0f, 4.239733f, 0.0f, 0.111106396f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -1.3735518f, 0.0f, 0.0f, 0.0f, 0.0f, 9.497997f, 0.0f},
						{0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.111106396f, 0.0f, 5.073994f}}
	};
	
}
