package reksten.xtree;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import reksten.base.FloatVector;

import neustore.base.DBRuntimeException;

public abstract class XTreeEntry<Data extends FloatVector> {
	
	static final short NODE_DIRECTORY = 1;
	static final short NODE_LEAF = 2;
	
	public static int SORTING_DIMENSION = 0;
	
	
	protected int nodeType;
	
	public abstract int size();
	
	public abstract int dimensionality();
	
	protected abstract void read(DataInputStream in) throws IOException;

	public abstract void updateParentMBR(XTreeDirectoryEntry<Data> parent);
	
//	public abstract boolean equals(XTreeEntry<Data> other);
//	
//	public abstract XTreeEntry<Data> copy();
	
	protected void write(DataOutputStream out) throws IOException {
		out.writeShort(this.nodeType);
	}
	
	@SuppressWarnings("unchecked")
	public static XTreeEntry parse(DataInputStream in) throws IOException {
		
		short nodeType = in.readShort();
		
		if(nodeType==NODE_LEAF) {
			XTreeLeafEntry leaf = new XTreeLeafEntry();
			leaf.read(in);
			return leaf;
		} else if(nodeType==NODE_DIRECTORY) {
			XTreeDirectoryEntry directory = new XTreeDirectoryEntry();
			directory.read(in);
			return directory;
		}
		System.out.println(nodeType);
		throw new DBRuntimeException("Node type does not exist!");
	}
	
}
