package reksten.xtree;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import reksten.base.DataObject;
import reksten.base.FloatVector;


public class XTreeLeafEntry<Data extends FloatVector> extends XTreeEntry<Data> implements Comparable<XTreeLeafEntry<Data>> {

	private Data object;
	
	XTreeLeafEntry() {this.nodeType = NODE_LEAF;}
	
	XTreeLeafEntry(Data object) {
		this.nodeType = XTreeEntry.NODE_LEAF;
		this.object = object;
	}
	
	public Data getObject() {
		return this.object;
	}
	
	@Override
	public int size() {
		int size = 2; // nodetype;
		size += object.size(); 
		return size;
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void read(DataInputStream in) throws IOException {
		this.object = (Data)DataObject.parse(in);
	}

	@Override
	protected void write(DataOutputStream out) throws IOException {
		super.write(out);
		object.write(out);
	}

	@Override
	public int dimensionality() {
		return this.object.dimensionality();
	}
	
	public static int SORTING_DIMENSION = 0;
	public int compareTo(XTreeLeafEntry<Data> arg0) {
		Data o1 = this.object, o2 = arg0.object;
		if(o1.get(SORTING_DIMENSION)<o2.get(SORTING_DIMENSION))
			return -1;
		if(o1.get(SORTING_DIMENSION)>o2.get(SORTING_DIMENSION))
			return 1;
		return 0;
	}

	@Override
	public void updateParentMBR(XTreeDirectoryEntry<Data> parent) {
		for(int i = 0;i<dimensionality();i++) {
			parent.mbr_low[i] = Math.min(parent.mbr_low[i],object.get(i));
			parent.mbr_high[i] = Math.max(parent.mbr_high[i],object.get(i));
		}
	}
	
}
